unit ufrmIPSXBuilder;

(*
              M3205 - ufrmIPSXBuilder
       (c) 2005 Paul C. L. Willmott, VP9MU

 This program is free software; you can redistribute it
 and/or modify it under the terms of the GNU General
 Public License as published by the Free Software
 Foundation; either version 2 of the License, or at
 your option, any later version.

 This program is distributed in the hope that it will
 be useful, but WITHOUT ANY WARRANTY; without even the
 implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public
 License for more details.

 You should have received a copy of the GNU General
 Public License along with this program; if not, write
 to the Free Software Foundation, Inc., 59 Temple
 Place, Suite 330, Boston, MA  02111-1307  USA

        Contact : pwillmott@northrock.bm
*)

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls;

const

  MaxImageSize = $1FFFF ;
  MaxStackSize = $400 ;
  HexChar : array[$0..$f] of char = ('0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F') ;

  I_DEFEX      = 0  ;
  I_VAREX      = 4  ;
  I_CONSEX     = 8  ;
  I_RETEX      = 12 ;
  I_4BLITERAL  = 16 ;
  I_BRONZ      = 20 ;
  I_JUMP       = 24 ;
  I__JEEX      = 28 ;
  I_LOOPEX     = 32 ;
  I_PLUSLOOPEX = 36 ;

type

  DictionaryEntryPtr = ^DictionaryEntry ;

  DictionaryEntry = record
    Identifier : string[63]         ;
    DictAddr   : Cardinal           ;
    NextEntry  : DictionaryEntryPtr ;
    LastEntry  : DictionaryEntryPtr ;
  end ; { DictionaryEntry }

  MacroDefinitionPtr = ^MacroDefinition ;

  MacroDefinition = record
    Identifier : string[63] ;
    Macro      : string[255] ;
    NextEntry  : MacroDefinitionPtr ;
  end ; { MacroDefinition }

  MacroBufferPtr = ^MacroBuffer ;

  MacroBuffer = record
    Macro      : string[255] ;
    NextEntry  : MacroBufferPtr ;
  end ; { MacroBuffer }

  TfrmIPSXBuilder = class(TForm)
    pgcntrlBuilder: TPageControl;
    tbshtEcho: TTabSheet;
    tbshtDump: TTabSheet;
    tbshtDictionary: TTabSheet;
    tbshtPCode: TTabSheet;
    memEcho: TRichEdit;
    memDump: TRichEdit;
    memDictionary: TRichEdit;
    memPCode: TRichEdit;
    dlgOpenFile: TOpenDialog;
    lblFilename: TLabel;
    edtFilename: TEdit;
    btnOpen: TButton;
    btnBuild: TButton;
    procedure btnBuildClick(Sender: TObject);
    procedure btnOpenClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    { Private declarations }

    MemImage    : array[0..MaxImageSize-1] of byte ; { Assembled IPS-X Memory Image                          }
    _H          : Cardinal ;                         { Pointer to Free Memory Position                       }
    _KK         : Cardinal ;                         { Address of Start of Current Dictionary Entry          }
    CompileFlag : boolean ;                          { True if compiling a Definition                        }
    EchoFlag    : boolean ;                          { True if Input Stream echoed during assembly           }
    InComment   : boolean ;                          { TRUE if inside a Comment                              }
    ASMFlag     : boolean ;                          { TRUE when in Assembler Mode, Numbers written to Image }
    Stack       : array[0..MaxStackSize-1] of Cardinal ; { Stack for Jump Addresses etc                      }
    StackPtr    : Cardinal ;                         { Pointer to Next Free Position on Stack                }
    Dictionary  : DictionaryEntryPtr ;               { Pointer to Start of Dictionary List                   }
    MacroList   : MacroDefinitionPtr ;               { Pointer to Start of Macro List                        }
    MacroStack  : MacroBufferPtr ;                   { Macro Input Buffer                                    }
    SourceFile  : file of char ;                     { Input Source Code File                                }
    OutFile     : file of byte ;                     { Memory Image File }
    EchoFile    : Text ;
    LogFile     : Text ;
    InWordStr   : string ;                           { Word from Source File to Process                      }
    InString    : BOOLEAN ;                          { TRUE if inside String Literal                         }
    sLiteral    : string ;                           { String Literal may be used once only                  }
    kk          : byte ;
    BaseName    : string ; { Base Filename, e.g. "IPS-MP" from "IPS-MP.SRC" }

    procedure CompileError(sError:string) ;
    procedure SaveSettings ;
    procedure PushCardinal(M:Cardinal) ;
    procedure PushInteger(M:integer) ;
    function PopCardinal:Cardinal ;
    function PopInteger:integer ;
    function PeekByte(Addr:Cardinal):Byte ;
    function PeekCardinal(Addr:Cardinal):Cardinal ;
    procedure PokeByte(Addr,Value:Cardinal) ;
    procedure PokeCardinal(Addr,Value:Cardinal) ;
    function GetDictAddr(sIdentifier:string):Cardinal ;
    function GetDictIdentifier(Addr:Cardinal):string ;
    function GetCCODE(Index:Cardinal):Cardinal ;
    function MacroWord(NewIdentifier:string;Test:boolean):boolean ;
    function GetByte(var cc:char):boolean ;
    function GetNextWord(DelimChar:char;var InWordStr:string):boolean ;
    procedure WriteToImageByte(bb:byte) ;
    procedure WriteToImageCardinal(ww:Cardinal) ;
    procedure WriteLiteral(ww:Cardinal) ;
    function NumericValue(InWordStr:string;Test:boolean):boolean ;
    function DefinedWord(InWordStr:string;Test:boolean):boolean ;
    function ReservedWord(InWordStr:string;Test:boolean):boolean  ;
    procedure EntrySetup(CodeAddr,EntryType:Cardinal) ;
    procedure DumpList(DPtr:DictionaryEntryPtr;Full:boolean) ;
    procedure UpdateDisplays ;
  public
    { Public declarations }
  end;
  function HexStrByte(bb:Cardinal):string ;
  function HexStrCardinal(ww:Cardinal):string ;

var
  frmIPSXBuilder: TfrmIPSXBuilder;

  function PolyName(S:string):Cardinal ;

implementation

uses ufrmM3205, StrUtils, INIFiles ;

{$R *.dfm}

procedure TfrmIPSXBuilder.btnBuildClick(Sender: TObject);
var
  iAddr : Cardinal ;
  DPtr  : DictionaryEntryPtr ;
  MLPtr : MacroDefinitionPtr ;
  MSPtr : MacroBufferPtr ;

begin

  SaveSettings ;

// Open Source File

  BaseName := edtFilename.Text ;

  AssignFile(SourceFile,BaseName) ;
  Reset(SourceFile) ;

  kk := Pos('.',BaseName) ;

  if (kk>0) then begin
    BaseName := LeftStr(BaseName,kk-1) ;
  end ;

  AssignFile(EchoFile,BaseName+' Echo List.txt') ;
  ReWrite(EchoFile) ;

// Blank Binary and Reporting Files

  AssignFile(OutFile,BaseName+'.BIN') ;
  ReWrite(OutFile) ;
  CloseFile(OutFile) ;
  AssignFile(LogFile,BaseName+' Memory Dump.txt') ;
  ReWrite(LogFile) ;
  CloseFile(LogFile) ;

  AssignFile(LogFile,BaseName+' Definition List.txt') ;
  ReWrite(LogFile) ;
  CloseFile(LogFile) ;

  AssignFile(LogFile,BaseName+' PCode List.txt') ;
  ReWrite(LogFile) ;
  CloseFile(LogFile) ;

//  Initialise Workspace

  StackPtr := MaxStackSize-1 ;

  FillChar(MemImage,SizeOf(MemImage),$00) ;

  InComment := false ;
  InString := false ;
  ASMFlag := false ;
  _H := $00000000 ;
  _KK := $00000000 ;
  Dictionary := nil ;
  MacroList := nil ;
  MacroStack := nil ;
  sLiteral := '' ;
  EchoFlag := false ;
  StackPtr := MaxStackSize - 1 ;

//  Process Source File

  while (not Eof(SourceFile)) do begin

//  Get Next Word, handling comments as necessary

    if (GetNextWord(' ',InWordStr)) then begin

//  Check for Macro Substitutions

      if (not MacroWord(InWordStr,false)) then begin

//  Check for IPSXBuilder Reserved Words

        if (not ReservedWord(InWordStr,false)) then begin

//  Check for Numbers

          if (not NumericValue(InWordStr,false)) then begin

//  Check for Previously Defined Words

            if (not DefinedWord(InWordStr,false)) then begin

//  Flag Error & Quit

              if (Trim(InWordStr)<>'') then begin
                CompileError('Unknown definition ['+Trim(InWordStr)+']') ;
              end ;

            end ;

          end ;

        end ;

      end ;

    end ;

  end ;

  if (StackPtr<>MaxStackSize-1) then begin
    CompileError('Items Left on Stack.') ;
  end ;

{
     Close Source File
}

  CloseFile(SourceFile) ;
  CloseFile(EchoFile) ;

{
     Check if Assembly Completed, and write Image if so!
}

  AssignFile(OutFile,BaseName+'.BIN') ;
  ReWrite(OutFile) ;
  AssignFile(LogFile,BaseName+' Memory Dump.txt') ;
  ReWrite(LogFile) ;

  for iAddr := $00000000 to _H do begin
    kk :=PeekByte(iAddr) ;
    Write(OutFile,kk) ;
    if (iAddr mod 16 = 0) then begin
      WriteLn(LogFile) ;
      Write(LogFile,'#' + HexStrCardinal(iAddr)+': ' ) ;
    end ;
    Write(LogFile,'#'+HexStrByte(PeekByte(iAddr))+' ') ;
  end ;
  WriteLn(LogFile) ;
  CloseFile(LogFile) ;
  CloseFile(OutFile) ;

  AssignFile(LogFile,BaseName+' Definition List.txt') ;
  ReWrite(LogFile) ;
  DumpList(Dictionary,false) ;
  CloseFile(LogFile) ;

  AssignFile(LogFile,BaseName+' PCode List.txt') ;
  ReWrite(LogFile) ;
  DumpList(Dictionary,true) ;
  CloseFile(LogFile) ;

{
     De-Allocate Virtual Memory Space
}

  while (Dictionary<>nil) do begin
    DPtr := Dictionary^.NextEntry ;
    Dispose(Dictionary) ;
    Dictionary := DPtr ;
  end ;
  while (MacroList<>nil) do begin
    MLPtr := MacroList^.NextEntry ;
    Dispose(MacroList) ;
    MacroList := MLPtr ;
  end ;
  while (MacroStack<>nil) do begin
    MSPtr := MacroStack^.NextEntry ;
    Dispose(MacroStack) ;
    MacroStack := MSPtr ;
  end ;

  UpdateDisplays ;

  ShowMessage('Build Complete!') ;

end ;

procedure TfrmIPSXBuilder.CompileError(sError: string);
begin
  WriteLn(EchoFile) ;
  WriteLn(EchoFile,sError) ;
  CloseFile(EchoFile) ;
  CloseFile(SourceFile) ;
  UpdateDisplays ;
  raise Exception.Create(sError) ;
end;

function TfrmIPSXBuilder.DefinedWord(InWordStr: string;
  Test: boolean): boolean;
{
     Defined Word Handler
}

var
  DictEntry : DictionaryEntryPtr ;
  GotWord   : boolean            ;
  nType     : Cardinal           ;
  nAddr     : Cardinal           ;
begin
  InWordStr := Trim(InWordStr) ;
  result := false ;
  GotWord := false ;
  DictEntry := Dictionary ;
  while (not GotWord) and (DictEntry<>nil) do begin
    if (InWordStr=DictEntry^.Identifier) then begin
      if (not Test) then begin
        nAddr := DictEntry^.DictAddr+8 {6} ;
        if (CompileFlag) then begin
          WriteToImageCardinal(nAddr) ;
        end
        else begin
          nType := PeekCardinal(nAddr) ;
          if (nType=GetCCODE(I_CONSEX)) then begin
            PushCardinal(PeekCardinal(nAddr+4)) ;
          end
          else if (nType=GetCCODE(I_VAREX)) then begin
            PushCardinal(nAddr+4) ;
          end
          else begin
            CompileError('Definition calls not allowed outside compile mode') ;
          end ;
        end ;
      end ;
      GotWord := true ;
    end ;
    DictEntry := DictEntry^.NextEntry ;
  end ;

  result := GotWord ;

end ; { DefinedWord }

procedure TfrmIPSXBuilder.DumpList(DPtr: DictionaryEntryPtr;
  Full: boolean);

var
  LL,AA,OP : Cardinal ;
  ss : string ;
  OpType : Cardinal ;

begin

  if (DPtr<>nil) then begin

     while (DPtr^.NextEntry<>nil) do begin
       DPtr := DPtr^.NextEntry ;
     end ;

     repeat

       if (DPtr^.LastEntry=nil) then begin
         LL := _H-1 ;
       end
       else begin
         LL := DPtr^.LastEntry^.DictAddr-1 ;
       end ;

       if (Full) then begin
         WriteLn(LogFile) ;
       end ;

       AA := DPtr^.DictAddr+8 ;
       OPType := PeekCardinal(AA) ;

       WriteLn(LogFile, '#' + HexStrCardinal(AA) + '  #'+HexStrCardinal(OPType)+'  [' + DPtr^.Identifier+']') ;
       AA := AA + 4 ;

       if (Full) then begin
         while (AA<=LL) do begin
           OP := PeekCardinal(AA) ;
           ss := GetDictIdentifier(OP) ;

           if (OpType=GetCCODE(I_DEFEX)) then begin

             if      (ss='4bliteral') or
                     (ss='jump') or
                     (ss='bronz') or
                     (ss='loopex') or
                     (ss='+loopex') or
                     (ss='$jeex')
                     then begin
               WriteLn(LogFile,'#'+HexStrCardinal(AA)+'  '+ss) ;
               AA := AA + 4 ;
               WriteLn(LogFile,'#'+HexStrCardinal(AA)+'  #'+HexStrCardinal(PeekCardinal(AA))) ;
               AA := AA + 4 ;
             end
             else if (ss='') then begin
               WriteLn(LogFile,'#'+HexStrCardinal(AA)+'  #'+HexStrCardinal(OP)) ;
               AA := AA + 4 ;
             end
             else begin
               WriteLn(LogFile,'#'+HexStrCardinal(AA)+'  '+ss) ;
               AA := AA + 4 ;
             end ;
           end
           else begin
             WriteLn(LogFile,'#'+HexStrCardinal(AA)+'  #'+HexStrByte(OP)) ;
             AA := AA + 1 ;
           end ;
         end ;

       end ;

       DPtr := DPtr^.LastEntry ;

     until (DPtr=nil) ;

  end ;

end ; { DumpList }

procedure TfrmIPSXBuilder.EntrySetup(CodeAddr, EntryType: Cardinal);
{
     Setup Dictionary Entry
}

var
  Hash      : Cardinal           ;
  DictEntry : DictionaryEntryPtr ;
  InWordStr : string             ;

begin

{
     Get the Entry Name
}

  if (not (GetNextWord(' ',InWordStr))) then begin
    CompileError('Syntax Error.') ;
  end ;

  InWordStr := Trim(InWordStr) ;

{
     Check its not already used
}

  if (MacroWord   (InWordStr,true)) or
     (ReservedWord(InWordStr,true)) or
     (DefinedWord (InWordStr,true)) then begin
    CompileError('The token ' + InWordStr + ' is a duplicate.') ;
  end ;

{
     Create New Dictionary Entry
}

  New(DictEntry) ;

  DictEntry^.NextEntry := Dictionary ;
  if (Dictionary<>nil) then begin
    Dictionary^.LastEntry := DictEntry ;
  end ;
  DictEntry^.LastEntry := nil ;
  Dictionary := DictEntry ;
  DictEntry^.Identifier := InWordStr ;
  Hash := PolyName(InWordStr) ;
  DictEntry^.DictAddr := _H ;

{
     Write to Memory Image
}

  WriteToImageByte(Length(InWordStr) or EntryType) ;
  WriteToImageByte(T4Byte(Hash).a) ;
  WriteToImageByte(T4Byte(Hash).b) ;
  WriteToImageByte(T4Byte(Hash).c) ;
  WriteToImageCardinal(_KK) ;
  WriteToImageCardinal(CodeAddr) ; // Code Addr now at +8 instead of +6

{
     Store Link for Next Entry
}

  _KK := DictEntry^.DictAddr ;

end ; { EntrySetup }

function TfrmIPSXBuilder.GetByte(var cc: char): boolean;
{
     This routine gets the next byte from the input stream; File or Macro Stack

     Returns TRUE if successful or FALSE if end of input is reached
}

var
  GotChar : boolean        ;
  tPtr    : MacroBufferPtr ;

begin

  GotChar := false ;

  if (MacroStack<>nil) then begin
    tPtr := MacroStack ;
    with tPtr^ do begin
      cc := Macro[1] ;                    // Get first character of string
      GotChar := true ;
      Move(Macro[2],Macro[1],Length(Macro)-1) ;
      SetLength(Macro,Length(Macro)-1) ;  // Delete 1st character of string
      if (Length(Macro)=0) then begin     // When the string is empty remove
                                          // entry from stack
        MacroStack := NextEntry ;
        Dispose(tPtr) ;
      end ;
    end ;
  end ;

  if (not GotChar) and (not Eof(SourceFile)) then begin
    Read(SourceFile,cc) ;
    if (not (cc in [' '..'~'])) then begin  // replace all control characters
                                            // with spaces }
      cc := ' ' ;
    end ;
    GotChar := true ;
  end ;
  if (EchoFlag) then begin
    Write(EchoFile,cc) ;
  end ;

  result := GotChar ;

end;

function TfrmIPSXBuilder.GetCCODE(Index: Cardinal): Cardinal;
var
  nAddr : Cardinal ;
begin
  nAddr := GetDictAddr('$ncodes')+12 {8} ; // 4 + 4 + 4  was 4 + 2 + 2
  if (nAddr=$00000000) then begin
    CompileError('The field $ncodes has not be defined in the source code.') ;
  end ;
  result := PeekCardinal(nAddr+Index) ;
end ; { GetCCODE }

function TfrmIPSXBuilder.GetDictAddr(sIdentifier: string): Cardinal;
var
  tPtr : DictionaryEntryPtr ;
begin
  sIdentifier := Trim(sIdentifier) ;
  result := $00000000 ;
  tPtr := Dictionary ;
  while (tPtr<>nil) and (tPtr^.Identifier<>sIdentifier) do begin
    tPtr := tPtr^.NextEntry ;
  end ;
  if (tPtr<>nil) then begin
    result := tPtr^.DictAddr ;
  end ;
end ; { GetDictAddr }

function TfrmIPSXBuilder.GetDictIdentifier(Addr: Cardinal): string;
var
  tPtr : DictionaryEntryPtr ;
begin
  result := '' ;
  tPtr := Dictionary ;
  while (tPtr<>nil) and (tPtr^.DictAddr+8<>Addr) do begin
    tPtr := tPtr^.NextEntry ;
  end ;
  if (tPtr<>nil) then begin
    result := tPtr^.Identifier ;
  end ;
end;

function TfrmIPSXBuilder.GetNextWord(DelimChar: char;
  var InWordStr: string): boolean;

{
     This routine gets a sequence of characters from the input stream upto and
     including the requested delimiter character.

     Returns the requested string in ss.

     The function returns TRUE if successful or FALSE if end of input is reached.
}

var
  GotToken  : boolean ;
  cc        : char ;
  Quit      : boolean ;

begin

  InWordStr := '' ;
  GotToken := false ;

  if (DelimChar=' ') then begin
    while (GetByte(cc)) and (cc=' ') do begin
    end ;
  end
  else begin
    if (GetByte(cc)) then ;
  end ;
  Quit := false ;
  while (not Quit) and (not GotToken) do begin
    InWordStr := InWordStr + cc ;
    if (cc=DelimChar) then begin
      GotToken := true ;
    end ;
    if (not GotToken) and (not (GetByte(cc))) then begin
      Quit := true ;
    end ;
  end ;

  if (Length(InWordStr)<>0) and (not (GotToken)) and (DelimChar=' ') then begin
    InWordStr := InWordStr + DelimChar ;
    GotToken := true ;
  end ;

  result := GotToken ;

end ; { GetNextWord }

function HexStrByte(bb: Cardinal): string;
begin
  bb := bb and $ff ;
  result := HexChar[bb shr 4] + HexChar[bb and $f] ;
end;

function HexStrCardinal(ww: Cardinal): string;
begin
  result := HexStrByte(T4Byte(ww).d)+
            HexStrByte(T4Byte(ww).c)+
            HexStrByte(T4Byte(ww).b)+
            HexStrByte(T4Byte(ww).a) ;
end ; { HexStrCardinal }

function TfrmIPSXBuilder.MacroWord(NewIdentifier: string;
  Test: boolean): boolean;
var
  TokenFound : boolean ;
  ListPtr    : MacroDefinitionPtr ;
  tPtr       : MacroBufferPtr ;
begin
  TokenFound := false ;
  ListPtr := MacroList ;
  NewIdentifier := Trim(NewIdentifier) ;
  while (not TokenFound) and (ListPtr<>nil) do begin
    if (ListPtr^.Identifier=NewIdentifier) then begin
      TokenFound := true ;
      if (not Test) then begin
        New(tPtr) ;
        tPtr^.NextEntry := MacroStack ;
        tPtr^.Macro := ListPtr^.Macro ;
        MacroStack := tPtr ;
        tPtr := nil ;
      end ;
    end ;
    ListPtr := ListPtr^.NextEntry ;
  end ;
  result := TokenFound ;
end ; { MacroWord }

function TfrmIPSXBuilder.NumericValue(InWordStr: string;
  Test: boolean): boolean;
var
  IsHex       : boolean ;
  Basis       : Int64 ;
  NumberGood  : boolean ;
  Digit       : Int64 ;
  Value       : Int64 ;
  Sign        : Int64 ;
  ii, jj      : integer ;
  WValue      : Cardinal ;
  IValue      : Integer ;

begin

  NumberGood := false ;

  InWordStr := Trim(InWordStr) ;

  Sign := +1 ;

  if (InWordStr<>'') then begin

    IsHex := (InWordStr[1]='#') ;

    if (IsHex) then begin
      Basis := 16 ;
      InWordStr := RightStr(InWordStr,Length(InWordStr)-1) ;
    end
    else begin
      if (InWordStr[1]='-') then begin
        Sign := -1 ;
        InWordStr := RightStr(InWordStr,Length(InWordStr)-1) ;
      end
      else begin
        Sign := +1 ;
      end ;
      Basis := 10 ;
    end ;

    if (InWordStr<>'') then begin
       Value := 0 ;
       for ii := 1 to Length(InWordStr) do begin
         Digit := -1 ;
         for jj := 0 to Basis-1 do begin
           if (InWordStr[ii]=HexChar[jj]) then begin
             Digit := jj ;
           end ;
         end ;
         if (Digit=-1) then begin
           Exit ;
         end
         else begin
           Value := Value * Basis + Digit ;
         end ;
       end ;
       if (IsHex) then begin
         if (Value<0) or (Value>$FFFFFFFF) then begin
           Exit ;
         end ;
         WValue := Value and $FFFFFFFF ;
         if (not Test) then begin
           WriteLiteral(WValue) ;
         end ;
       end
       else begin
         Value := Value * Sign ;
         if (Value<(-2147483647-1) ) or (Value>(2147483646+1)) then begin
           Exit ;
         end ;
         IValue := Value ;
         if (not Test) then begin
           WriteLiteral(Cardinal(IValue)) ;
         end ;
       end ;

       NumberGood := true ;

     end ;

   end ;

   result := NumberGood ;

end ; { NumericValue }

function TfrmIPSXBuilder.PeekByte(Addr: Cardinal): byte;
begin
  result := MemImage[Addr] ;
end;

function TfrmIPSXBuilder.PeekCardinal(Addr: Cardinal): Cardinal;
begin
  T4Byte(result).a := PeekByte(Addr+0) ;
  T4Byte(result).b := PeekByte(Addr+1) ;
  T4Byte(result).c := PeekByte(Addr+2) ;
  T4Byte(result).d := PeekByte(Addr+3) ;
end;

procedure TfrmIPSXBuilder.PokeByte(Addr, Value: Cardinal);
begin
  if (Addr>MaxImageSize-1) then begin
    CompileError('Write to memory outside of Image Buffer attempted.') ;
  end ;
  MemImage[Addr] := Lo(Value) ;
end;

procedure TfrmIPSXBuilder.PokeCardinal(Addr, Value: Cardinal);
begin
  PokeByte(Addr+0 ,T4Byte(Value).a) ;
  PokeByte(Addr+1 ,T4Byte(Value).b) ;
  PokeByte(Addr+2 ,T4Byte(Value).c) ;
  PokeByte(Addr+3 ,T4Byte(Value).d) ;
end;

function PolyName(S: string): Cardinal;

{
     REFERENCE:  IPS High Level Programming of Small Systems, 2nd Edition, Karl Meinzer, 1978
                 Page 76 - 5.2.1 Scanner and name coding
                 Page 81 - 5.4 Entry generating routines ($POLYNAME examaple!!!!)
}

var
  D : Cardinal ;
  X : Cardinal ;
  P : Cardinal ;
  I : byte ;

begin

  X := $57ce8d ; //  "VERT" encodes as #EFBD72
 
  for I := 1 to Length(S) do begin
    D := Ord(S[I]) ;
    P := D xor X xor (X shr 1) xor (X shr 2) xor (X shr 7) ;
    T4Byte(X).d := T4Byte(P).a ;
    X := X shr 7 ;
  end ;

  result := X and $ffffff ;

end ; { PolyName }

function TfrmIPSXBuilder.PopCardinal: Cardinal;
begin
  if (StackPtr=MaxStackSize-1) then begin
    CompileError('Stack Underflow') ;
  end ;
  Inc(StackPtr) ;
  result := Stack[StackPtr] ;
end;

function TfrmIPSXBuilder.PopInteger: integer;
begin
  result := integer(PopCardinal) ;
end;

procedure TfrmIPSXBuilder.PushCardinal(M: Cardinal);
begin
  if (StackPtr=0) then begin
    CompileError('Stack Overflow') ;
  end ;
  Stack[StackPtr] := M ;
  Dec(StackPtr) ;
end;

procedure TfrmIPSXBuilder.PushInteger(M: integer);
begin
  PushCardinal(Cardinal(M)) ;
end;

function TfrmIPSXBuilder.ReservedWord(InWordStr: string;
  Test: boolean): boolean;

var
  NoneFound       : boolean ;
  tStr            : string ;
  MacroDef        : MacroDefinitionPtr ;
  sID             : string ;
  cc              : char ;
  sTag            : string ;
  nAddr           : Cardinal ;
  nVal            : Cardinal ;
  SourceAddr      : Cardinal ;
  DestinationAddr : Cardinal ;
  Count           : Cardinal ;
  M               : integer ;

begin

  NoneFound := false ;

  InWordStr := Trim(InWordStr) ;

{
     Gobble Up Comments
}

  if (InWordStr='(') then begin
    sTag := InWordStr+' ' ;
    while (sTag<>' )') and (GetByte(cc)) do begin
      sTag[1] := sTag[2] ;
      sTag[2] := cc ;
    end ;
    InComment := (sTag = ' )') ;
  end

{
     Handle Macro Definitions
}

  else if (InWordStr=':MACRO') then begin

    if (not Test) then begin

      if (not ((GetNextWord(' ',InWordStr)) and (Trim(InWordStr)<>';MACRO'))) then begin
        CompileError('Syntax Error.') ;
      end ;

      sID := Trim(InWordStr) ;

      if (MacroWord(sID,true)) or
         (ReservedWord(sID,true)) or
         (DefinedWord(sID,true)) then begin
        CompileError('The Macro name ' + InWordStr + ' is already used.') ;
      end ;

      tStr := ' ' ;

      while ((GetNextWord(' ',InWordStr)) and (Trim(InWordStr)<>';MACRO')) do begin
        InWordStr := Trim(InWordStr) ;

        if (not (MacroWord(InWordStr,true))) and
           (not (ReservedWord(InWordStr,true))) and
           (not (DefinedWord(InWordStr,true))) and
           (not (NumericValue(InWordStr,true))) then begin
          CompileError('The token ' + InWordStr + ' is not numeric or is unknown.') ;
        end ;

        tStr := tStr + InWordStr + ' ' ;

      end ;

      if (Trim(InWordStr)=';MACRO') then begin
        New(MacroDef) ;
        with MacroDef^ do begin
          Identifier := sID ;
          Macro := tStr ;
          NextEntry := MacroList ;
        end ;
        MacroList := MacroDef ;
      end
      else begin
        CompileError('Macro definition not completed.') ;
      end ;

    end ;

  end

{
     Store Memory Pointer
}

  else if (InWordStr='!$H') then begin
    if (not Test) then begin
      _H := PopCardinal ;
    end ;
  end

  else if (InWordStr=',') then begin
    if (not Test) then begin
      PokeByte(_H,PopCardinal) ;
      Inc(_H) ;
    end ;
  end

  else if (InWordStr='@LINK') then begin
    if (not Test) then begin
      PushCardinal(_KK) ;
    end ;
  end

  else if (InWordStr='H4INC') then begin
    if (not Test) then begin
      Inc(_H,4) ;
    end ;
  end

  else if (InWordStr='$DEPW') then begin
    if (not Test) then begin
      PokeCardinal(_H,PopCardinal) ;
      Inc(_H,4) ;
    end ;
  end

  else if (InWordStr='INCR') then begin
    if (not Test) then begin
      nAddr := PopCardinal ;
      PokeCardinal(nAddr,Succ(PeekCardinal(nAddr))) ;
    end ;
  end

  else if (InWordStr='KON') then begin
    if (not Test) then begin
      EntrySetup(GetCCODE(I_CONSEX),0) ;
      WriteToImageCardinal(PopCardinal) ;
    end ;
  end

  else if (InWordStr='VAR') then begin
    if (not Test) then begin
      EntrySetup(GetCCODE(I_VAREX),0) ;
      WriteToImageCardinal(PopCardinal) ;
    end ;
  end

  else if (InWordStr='FELD') then begin
    if (not Test) then begin
      EntrySetup(GetCCODE(I_VAREX),0) ;
      Inc(_H,PopCardinal) ;
      _H := (_H + 3) and $fffffffc ;
    end ;
  end

  else if (InWordStr='RCODE') then begin
    if (not Test) then begin
      _H := (_H + 3) and $fffffffc ;
      EntrySetup(PopCardinal,0) ;
    end ;
  end

  else if (InWordStr=':') then begin
    if (not Test) then begin
      EntrySetup(GetCCODE(I_DEFEX),0) ;
      CompileFlag := true ;
    end ;
  end

  else if (InWordStr=':HPRI') then begin
    if (not Test) then begin
      EntrySetup(GetCCODE(I_DEFEX),$40) ;
      CompileFlag := true ;
    end ;
  end

  else if (InWordStr=':PRIOR') then begin
    if (not Test) then begin
      EntrySetup(GetCCODE(I_DEFEX),$80) ;
      CompileFlag := true ;
    end ;
  end

  else if (InWordStr=':INT') then begin
    if (not Test) then begin
      EntrySetup(GetCCODE(I_DEFEX),$C0) ;
      CompileFlag := true ;
    end ;
  end

{
     Turn Echo On
}

  else if (InWordStr='ECHO-ON') then begin
    if (not Test) then begin
      EchoFlag := true ;
    end ;
  end

{
     Turn Echo Off
}

  else if (InWordStr='ECHO-OFF') then begin
    if (not Test) then begin
      EchoFlag := false ;
    end ;
  end

{
     Exit Compile Mode
}

  else if (InWordStr='A>') then begin
    if (not Test) then begin
      CompileFlag := false ;
    end ;
  end

{
     Enter Compile Mode
}

  else if (InWordStr='<A') then begin
    if (not Test) then begin
      CompileFlag := true ;
    end ;
  end

{
     Modify Header
}

  else if (InWordStr='PRIMODIFY') then begin
    if (not Test) then begin
      PokeByte(_KK,PeekByte(_KK) or Lo(PopCardinal)) ;
    end ;
  end

{
     Push Memory Pointer to Stack
}

  else if (InWordStr='HIER') then begin
    if (not Test) then begin
      PushCardinal(_H) ;
    end ;
  end

{
     Store in Image
}

  else if (InWordStr='!') then begin
    if (not Test) then begin
      nAddr := PopCardinal ;
      PokeCardinal(nAddr,PopCardinal) ;
    end ;
  end

{
     Store Byte in Image
}

  else if (InWordStr='!B') then begin
    if (not Test) then begin
      nAddr := PopCardinal ;
      PokeByte(nAddr,PopCardinal) ;
    end ;
  end

{
     Get From Image
}

  else if (InWordStr='@') then begin
    if (not Test) then begin
      PushCardinal(PeekCardinal(PopCardinal)) ;
    end ;
  end

{
     Get Byte From Image
}

  else if (InWordStr='@B') then begin
    if (not Test) then begin
      PushCardinal(PeekByte(PopCardinal)) ;
    end ;
  end

{
     Store Field Components
}

  else if (InWordStr='!FK') then begin

    if (not Test) then begin

       nVal := PopCardinal ;
       nAddr := PopCardinal ;

       while (nVal<>0) do begin
         PokeCardinal(nAddr+(nVal-1)*4, PopCardinal) ;
         Dec(nVal) ;
       end ;

     end ;

  end

{
     Swap top 2 stack entries
}

  else if (InWordStr='VERT') then begin
    if (not Test) then begin
      nAddr := PopCardinal ;
      nVal := PopCardinal ;
      PushCardinal(nAddr) ;
      PushCardinal(nVal) ;
    end ;
  end

{
     Duplicate top of stack
}

  else if (InWordStr='DUP') then begin
    if (not Test) then begin
      nVal := PopCardinal ;
      PushCardinal(nVal) ;
      PushCardinal(nVal) ;
    end ;
  end

{
     Field Translate Long
}

  else if (InWordStr='>>>') then begin

    if (not Test) then begin

      Count           := PopCardinal ;
      DestinationAddr := PopCardinal ;
      SourceAddr      := PopCardinal ;

      if (Count>0) then begin // 1 <= Count <= 2147483647

        repeat
          PokeByte(DestinationAddr,MemImage[SourceAddr]) ;
          Inc(SourceAddr) ;
          Inc(DestinationAddr) ;
          Dec(Count) ;
        until (Count = 0) ;

      end ;

    end ;

  end

{
     +   Adds 2 2s complement words.
     <n> <m> +
     <n+m>

     REFERENCE: IPS Language Reference Manual Draft E 15 Feb 1996
}

  else if (InWordStr='+') then begin
    if (not Test) then begin
      PushInteger(PopInteger+PopInteger) ;
    end ;
  end

{
     -   Subtracts 2 2s complement words
     <n> <m> -
     <n-m>

     REFERENCE: IPS Language Reference Manual Draft E 15 Feb 1996
}

  else if (InWordStr='-') then begin
    if (not Test) then begin
      M := PopInteger ;              { <- variable used to guarentee execution order! }
      PushInteger(PopInteger-M) ;
    end ;
  end
{
     Get Address of Dictionary Entry Executable Code
}

  else if (InWordStr='''') then begin
    if (not Test) then begin
      if (not GetNextWord(' ',InWordStr)) then begin
        CompileError('Token expected.') ;
      end ;
      nVal := GetDictAddr(InWordStr) ;
      if (nVal=$00000000) then begin
        CompileError('The token ' + InWordStr + ' is unknown.') ;
      end ;
      Inc(nVal,8 {6}) ;
      WriteLiteral(nVal) ;
    end ;
  end
{
     String Literal
}

  else if (InWordStr='"') then begin
    if (not Test) then begin
      if (not GetNextWord('"',sLiteral)) then begin
        CompileError('Incomplete String Literal.') ;
      end ;
    end ;
  end

{
     Store String Literal
}

  else if (InWordStr='!T') then begin
    if (not Test) then begin
      nAddr := PopCardinal ;
      for M := 1 to Length(sLiteral)-2 do begin
        PokeByte(nAddr,byte(sLiteral[M])) ;
        Inc(nAddr) ;
      end ;
      sLiteral := '' ;
    end ;
  end

{
     Echo String Literal
}

  else if (InWordStr='ECHO') then begin
    if (not Test) and (sLiteral<>'') then begin
      WriteLn(EchoFile) ;
      WriteLn(EchoFile,LeftStr(sLiteral,Length(sLiteral)-2)) ;
      sLiteral := '' ;
    end ;
  end

{
     Set None Found Flag
}

  else begin
    NoneFound := true ;
  end ;

  result := not NoneFound ;

end ; { ReservedWord }

procedure TfrmIPSXBuilder.WriteLiteral(ww: Cardinal);
{
     This routine writes a numeric literal to the Image
}

begin
  if (CompileFlag) then begin
    WriteToImageCardinal(GetCCODE(I_4BLITERAL)) ;
    WriteToImageCardinal(ww) ;
  end
  else begin
    PushCardinal(ww) ;
  end ;
end ; { WriteLiteral }

procedure TfrmIPSXBuilder.WriteToImageByte(bb: byte);
{
     This routine writes a byte to the memory image
     and increments the memory pointer
}

begin
  PokeByte(_H,bb) ;
  Inc(_H) ;
end ; { WriteToImageByte }

procedure TfrmIPSXBuilder.WriteToImageCardinal(ww: Cardinal);
{
     This routine writes a word to the memory image
     and increments the memory pointer as necessary
}

begin
  WriteToImageByte(T4Byte(ww).a) ;
  WriteToImageByte(T4Byte(ww).b) ;
  WriteToImageByte(T4Byte(ww).c) ;
  WriteToImageByte(T4Byte(ww).d) ;
end ; { WriteToImageCardinal }

procedure TfrmIPSXBuilder.btnOpenClick(Sender: TObject);
begin
  dlgOpenFile.InitialDir := edtFilename.Text ;
  dlgOpenFile.FileName := edtFilename.Text ;
  dlgOpenFile.Execute ;
  edtFilename.Text := dlgOpenFile.FileName ;
  SaveSettings ;
end;

procedure TfrmIPSXBuilder.FormShow(Sender: TObject);
var
  pINIFile : TINIFile ;
begin
  pINIFile := TIniFile.Create( ChangeFileExt( Application.ExeName, '.INI' ) );
  try
    edtFilename.Text := pINIFile.ReadString( 'Builder', 'Filename', 'c:\ips32x-m.src' );
  finally
    pIniFile.Free;
  end;
end;

procedure TfrmIPSXBuilder.SaveSettings;
var
  pINIFile : TINIFile ;
begin
  pINIFile := TIniFile.Create( ChangeFileExt( Application.ExeName, '.INI' ) );
  try
    pINIFile.WriteString( 'Builder', 'Filename', edtFilename.Text );
    pINIFile.UpdateFile ;
  finally
    pIniFile.Free;
  end;
end;

procedure TfrmIPSXBuilder.UpdateDisplays;

procedure Update(memMemo:TRichEdit;Filename:string) ;
var
  ss : string ;
  LogFile : file of char ;
  cc : char ;
begin
  AssignFile(LogFile,Filename) ;
  Reset(LogFile) ;
  ss := '' ;
  while (not Eof(LogFile)) do begin
    Read(LogFile,cc) ;
    ss := ss + cc ;
  end ;
  CloseFile(LogFile) ;
  memMemo.Lines.Add(ss) ;
end ; { Update }
begin
  Update(memEcho,BaseName+' Echo List.txt') ;
  Update(memDump,BaseName + ' Memory Dump.txt') ;
  Update(memDictionary,BaseName + ' Definition List.txt') ;
  Update(memPCode,BaseName + ' PCode List.txt') ;
end ; { UpdateDisplays }

end.
